//redFrik - released under gnu gpl license

//--changes 080206
//fixed the www links in the helpfile
//added jukebox example

//--changes 070827
//importing multichannel soundfiles is now allowed. but ony takes 1st channel.

//--changes 070811
//added more magic strings so now 32ch fasttracker should load
//added standardizePath here and there
//bugfix: now checks for slots with empty samples
//changed the synthdef so that one can change out bus while playing
//new ex.11 in help

//--todo:
//ripPattern (how to translate into say a pbind?)
//.next method that step through slots


//--a class that can parse & play back .mod files
RedMOD {
	classvar
		magicdict,
		<>sampleRate= 7159090.5;					//7159090.5 for NTSC, 7093789.2 for PAL
	var
		<isPaused,
		<isPrepared,								//flag true if server booted, samples loaded and synths sent
		clk,										//tempoclock. passed in or created internally
		<server,
		<outBusses,								//array with bus indexes for output
		<>muteTracks= #[],							//array with track indexes to mute
		<>index= 0,
		<>row= 0,
		<>tick= 0,
		<>patterns,								//array with RedMODPattern objects
		<>samples,								//array with 31 RedMODSample objects
		<name,
		<>order,
		<magic,
		<format,
		<numTracks,
		task,
		synths,									//array with one synth for each track
		filters,									//array with one filter for each track
		customSpeed,								//integer that when not nil overrides mod speed
		speed= 6;									//integer global speed
	
	*initClass {
		magicdict= (
			'M.K.' : #[4, "Protracker 4 channel"],
			'M!K!' : #[4, "Protracker 4 channel"],
			'FLT4' : #[4, "Startracker 4 channel"],
			'FLT8' : #[8, "Startracker 8 channel"],
			'2CHN' : #[2, "Fasttracker 2 channel"],
			'4CHN' : #[4, "Fasttracker 4 channel"],
			'6CHN' : #[6, "Fasttracker 6 channel"],
			'8CHN' : #[8, "Fasttracker 8 channel"],
			'10CH' : #[10, "Fasttracker 10 channel"],
			'12CH' : #[12, "Fasttracker 12 channel"],
			'14CH' : #[14, "Fasttracker 14 channel"],
			'16CH' : #[16, "Fasttracker 16 channel"],
			'18CH' : #[18, "Fasttracker 18 channel"],
			'20CH' : #[20, "Fasttracker 20 channel"],
			'22CH' : #[22, "Fasttracker 22 channel"],
			'24CH' : #[24, "Fasttracker 24 channel"],
			'26CH' : #[26, "Fasttracker 26 channel"],
			'28CH' : #[28, "Fasttracker 28 channel"],
			'30CH' : #[30, "Fasttracker 30 channel"],
			'32CH' : #[32, "Fasttracker 32 channel"],
			'CD81' : #[8, "Atari oktalyzer 8 channel"],
			'OKTA' : #[8, "Atari oktalyzer 8 channel"],
			'16CN' : #[16, "Taketracker 16 channel"],
			'32CN' : #[32, "Taketracker 32 channel"],
			'    ' : #[4, "15-instrument 4 channel"]
		);
	}
	*read {|path|
		^super.new.read(path)
	}
	
	//--convert between midinotes & amiga periods (8363.42= amiga samplerate)
	*period2note {|period|
		^(sampleRate/(period*2)/8363.42).ratiomidi.round(0.001)
	}
	*note2period {|note|
		^(sampleRate/(note.midiratio*2)/8363.42).round
	}
	
	//--check if file exists & then parse
	read {|path|
		path= path.standardizePath;
		if(File.exists(path), {
			this.prRead(path);
			isPrepared= false;
		}, {("RedMOD: file"+path+"not found").error});
	}
	
	//--load sample data into buffers & send required synthdefs
	prepareForPlay {|group, action|
		if(isPrepared.isNil, {
			"RedMOD: read a file first".error;
			^this
		});
		if(isPrepared.not, {
			group= group ? Server.default.defaultGroup;
			server= group.server;					//get server from group
			Routine.run{
				var halt= Condition.new;
				server.bootSync(halt);				//boot the server
				
				SynthDef(\redMODplayer, {|out= 0, bufnum, t_trig= 0, loop= 0, loopstart= 0, loopend= 0, rate= 0, vol= 1, pan= 0, interp= 2, sr= 8363.42|
					var len= BufFrames.kr(bufnum)-1;
					var pbr= rate*(sr/SampleRate.ir);
					var i= RedPhasor.ar(t_trig, pbr, 0, len, loop, loopstart, loopend);
					var z= BufRd.ar(1, bufnum, i, 0, interp);
					Out.ar(out, Pan2.ar(LeakDC.ar(z), pan, vol));
				}, #[0, 0, \tr, 0, 0, 0, 0, 0, 0, 0, 0]).store;
				
				SynthDef(\redMODfilter, {|out= 0, in= 0, mix= 0|
					var dry= In.ar(in, 2);
					var wet= LPF.ar(dry, 3300);		//amiga filter wannabe
					ReplaceOut.ar(out, XFade2.ar(dry, wet, mix*2-1));
				}, #[\ir, \ir, 0.05]).store;
				
				samples.do{|x| if(x.empty.not, {x.load(server)})};
				server.sync(halt);					//wait for all buffers to load
				isPrepared= true;
				task= Task{
					this.prPlay;					//this is what you override
				};
				action.value;
			}
		}, {"RedMOD: already prepared - just play".warn});
	}
	
	//--playback methods
	play {|out= 0, clock, quant= 0|
		if(isPrepared.not, {
			"RedMOD: must .prepareForPlay first".warn;
			^this;
		});
		if(out.isSequenceableCollection, {
			outBusses= out;
		}, {
			outBusses= out.dup(numTracks);
		});
		if(clock.isKindOf(TempoClock), {
			clk= clock;
			CmdPeriod.doOnce({this.stop});
		}, {
			clk= TempoClock(125/60);				//125 bpm is standard .mod default
			clk.permanent= true;
		});
		isPaused= false;
		customSpeed= nil;							//following gets reset when .play
		muteTracks= #[];
		speed= 6;									//speed 6 is standard .mod default
		index= 0;
		row= 0;
		tick= 0;
		task.play(clk, true, quant);
	}
	stop {
		task.stop;
		synths.do{|x| x.free};
		filters.do{|x| x.free};
	}
	pause {
		isPaused= true;
		task.pause;
	}
	resume {
		isPaused= false;
		task.resume;
	}
	save {|path|
		this.prSave(path.standardizePath);
	}
	free {
		isPrepared= false;
		if(this.isPlaying, {this.stop});
		samples.do{|x| x.free};
	}
	isPlaying {
		^task.isPlaying
	}
	clock {
		^clk
	}
	speed_ {|val|
		customSpeed= val
	}
	speed {
		^customSpeed ? speed
	}
	revertSpeed {
		customSpeed= nil
	}
	
	//--private
	prRead {|path|								//decode file & create samps, patts, slots
		var length, samplesLengths;
		var file= File(path, "r");
		file.seek(1080, 0);						//jump forward to check if correct mod
		magic= {file.getChar}.dup(4).join;			//read magic string
		//magic.postln;							//debug
		if(magicdict[magic.asSymbol].isNil, {			//make sure this is a proper .mod file
			file.close;
			"RedMOD: format not supported".error;
			^this
		});
		
		//--read header & create sample objects
		file.seek(0, 0);							//jump back to beginning of file
		name= {file.getChar}.dup(20).join;			//read module title
		samplesLengths= 0.dup(31);					//temp storage
		samples= {|i|
			var name= {file.getChar}.dup(22)			//read sample name
				.select{|x| x.ascii!=0}.join;		//& remove zero padding
			var len= file.getInt16*2;				//read sample length
			var tun= file.getInt8;					//read fine tuning
			var vol= file.getInt8;					//read volume
			var loopstart= file.getInt16*2;			//read loop start
			var looplen= file.getInt16*2;			//read loop length
			samplesLengths[i]= len;					//to know later how many bytes to read
			if(tun>7, {tun= tun-16});				//signed value for fine tuning
			if(len>0, {							//check for empty sample slots
				RedMODSample(name, tun, vol, loopstart, looplen, 8363.42);
			}, {
				RedMODEmptySample(name, tun, vol, loopstart, looplen, 8363.42);
			});
		}.dup(31);
		
		length= file.getInt8;						//read songlength -use order.size instead
		file.getInt8;								//-skip unused
		order= {file.getInt8}.dup(128).keep(length);	//read which order to play patterns (form)
		4.do{file.getChar};						//-skip magic string, done above
		
		//--figure out how many tracks & patterns
		numTracks= magicdict[magic.asSymbol][0];
		format= magicdict[magic.asSymbol][1];
		
		//--read patterns & create pattern objects
		patterns= {
			RedMODPattern({
				{
					var t= file.getInt32;			//read the 4 bytes
					var a= (t&2r11111111000000000000000000000000).rightShift(24);
					var b= (t&2r00000000111111110000000000000000).rightShift(16);
					var c= (t&2r00000000000000001111111100000000).rightShift(8);
					var d= (t&2r00000000000000000000000011111111);
					var i= (a&2r11110000)+(c&2r11110000).rightShift(4);
					var samp= samples[i-1];			//sample object, can be nil too
					var peri= (a&0x0F*0x100)+b;		//frequency in amiga periods
					var efxn= c&0x0F;				//effect number
					var efxp= d;					//effect parameter
					RedMODSlot(samp, peri, efxn, efxp);
				}.dup(numTracks);
			}.dup(64));
		}.dup(order.maxItem+1);
		
		//--read data into samples objects
		samples.do{|x, i|
			if(x.empty.not, {
				x.data= {
					file.getInt8+0.5/127.5;			//convert int8 to float. but keep sr 8363.42
				}.dup(samplesLengths[i]);
			});
		};
		
		file.close;
	}
	
	//--write the thing back to disk
	prSave {|path|
		var file;
		file= File(path, "w");
		file.write(name.extend(20, 0.asAscii));
		samples.do{|x, jjj|
			file.write(x.name.extend(22, 0.asAscii));
			file.putInt16(x.data.size*(8363.42/x.sampleRate)/2);
			file.putInt8(x.tuning);
			file.putInt8(x.volume);
			file.putInt16(x.loopstart/2);
			file.putInt16(x.looplen/2);
		};
		file.putInt8(order.size);
		file.putInt8(0);
		order.extend(128, 0).do{|x| file.putInt8(x)};
		file.write(magic.asString);
		patterns.do{|x|
			x.slots.do{|y|
				y.do{|z, j|
					var i= if(z.sample.notNil, {samples.indexOf(z.sample)+1}, {0});
					var a= i&2r11110000;
					var b= z.period.rightShift(8);
					var c= z.period&0xFF;
					var d= i&2r00001111;
					var e= z.efxindex;
					var f= z.efxparam;
					file.putInt8(a+b);
					file.putInt8(c);
					file.putInt8(d.leftShift(4)+e);
					file.putInt8(f);
				};
			};
		};
		samples.do{|x|
			if(x.empty.not, {
				if(x.sampleRate!=8363.42, {
					x.data.resamp1(x.data.size*(8363.42/x.sampleRate)).do{|y|
						file.putInt8(y*127.5-0.5);
					};
				}, {
					x.data.do{|y|
						file.putInt8(y*127.5-0.5);
					};
				});
			});
		};
		file.close;
	}
	
	//--a 'correct' playback routine with all effects implemented
	//  override this in subclasses or copy&paste into a function
	//  the reason it is so long is that it should be fairly easy
	//  to rewrite as a single function. see helpfile for example
	prPlay {
		var tracks;								//current slots (usually 4)
		var trackLoopStart= 0.dup(numTracks);
		var trackLoopEnd= 0.dup(numTracks);
		var trackVolume= 0.dup(numTracks);
		var trackPeriod= 0.dup(numTracks);			//amiga periods
		var efxPortaSpeed= 0.dup(numTracks);
		var efxPortaPeriod= 0.dup(numTracks);			//amiga periods
		var efxVibratoSpeed= 0.dup(numTracks);
		var efxVibratoDepth= 0.dup(numTracks);
		var efxVibratoCount= 0.dup(numTracks);
		var efxTremoloSpeed= 0.dup(numTracks);
		var efxTremoloDepth= 0.dup(numTracks);
		var efxTremoloCount= 0.dup(numTracks);
		var efxPatternBreakFlag;
		var efxGlissandoFlag= false;
		var efxVibratoWave= 0;
		var efxPatternLoopRow= 0;
		var efxPatternLoopCnt= 0;
		var efxTremoloWave= 0;
		var efxDelayNoteFlag;
		var efxDelayPattern= 0;
		var funcVibrato, funcVolumeSlide, funcPorta;
		
		//--set up synths & filters for each track
		filters= outBusses.collect{|x|				//amiga filers wannabe (usually 4)
			Synth(\redMODfilter, [\in, x, \out, x]);
		};
		synths= outBusses.collect{|x|				//one synth per track (usually 4)
			Synth(\redMODplayer, [\out, x]);
		};
		
		//--extra effect functions for the onces called twice
		funcVibrato= {|slot, i|
			var vib= 0;
			var efx= efxVibratoWave%4;
			if(efx==3, {efx= 3.rand});				//3 means to pick a random wave
			switch(efx,
				0, {								//vibrato sine wave
					vib= sin(efxVibratoCount[i]%64/64*2pi);
				},
				1, {								//vibrato ramp down
					vib= efxVibratoCount[i]%64/64*2-1;
				},
				2, {								//vibrato square wave
					vib= (efxVibratoCount[i]%64/64).round*2-1;
				}
			);
			vib= vib*efxVibratoDepth[i]*2;
			efxVibratoCount[i]= efxVibratoCount[i]+efxVibratoSpeed[i];
			synths[i].set(\rate, RedMOD.period2note(trackPeriod[i]+vib).midiratio);
		};
		funcVolumeSlide= {|slot, i|
			if(slot.efxparam.rightShift(4)>0, {
				trackVolume[i]= (trackVolume[i]+slot.efxparam.rightShift(4)).min(64);
				synths[i].set(\vol, trackVolume[i]/64);
			}, {
				if(slot.efxparam&15>0, {
					trackVolume[i]= (trackVolume[i]-(slot.efxparam&15)).max(0);
					synths[i].set(\vol, trackVolume[i]/64);
				});
			});
		};
		funcPorta= {|slot, i|
			if(trackPeriod[i]<efxPortaPeriod[i], {
				if(efxGlissandoFlag, {
					trackPeriod[i]= RedMOD.note2period(RedMOD.period2note(trackPeriod[i])+1);
				}, {
					trackPeriod[i]= (trackPeriod[i]+efxPortaSpeed[i]).min(efxPortaPeriod[i]);
				});
			}, {
				if(efxGlissandoFlag, {
					trackPeriod[i]= RedMOD.note2period(RedMOD.period2note(trackPeriod[i])-1);
				}, {
					trackPeriod[i]= (trackPeriod[i]-efxPortaSpeed[i]).max(efxPortaPeriod[i]);
				});
			});
			synths[i].set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
		};
		
		//--main loop
		while({index<order.size}, {					//loop ticks
			
			if(tick==0, {							//slots update every 0 tick, rest sub efxs
				efxPatternBreakFlag= false;
				efxDelayNoteFlag= false.dup(numTracks);
				tracks= patterns[order[index]].slots[row];
				tracks.do{|slot, i|				//for each track (usually 4)
					var trackSynth= synths[i];		//find corresponding synth object
					server.makeBundle(nil, {
						if(muteTracks.includes(i).not, {
							//--update synth if sample given & not empty
							if(slot.sample.notNil and:{slot.sample.empty.not}, {
								trackSynth.set(\bufnum, slot.sample.buffer.bufnum);
								trackSynth.set(\sr, slot.sample.sampleRate);
								trackSynth.set(\loop, slot.sample.loop.binaryValue);
								trackLoopStart[i]= slot.sample.loopstart;
								trackSynth.set(\loopstart, trackLoopStart[i]);
								trackLoopEnd[i]= slot.sample.loopend;
								trackSynth.set(\loopend, slot.sample.loopend);
								trackVolume[i]= slot.sample.volume;
							});
							if(slot.period>0, {	//update track period if note given
								if(slot.efxindex!=3 and:{slot.efxindex!=5}, {
									if(slot.sample.notNil, {
										trackPeriod[i]= slot.tunedPeriod;
										trackSynth.set(\rate, slot.tunedNote.midiratio);
									}, {
										trackPeriod[i]= slot.period;
										trackSynth.set(\rate, slot.note.midiratio);
									});
								});
							});
						});
						
						//--non tick based effects
						if((slot.efxindex==0 and:{slot.efxparam==0}).not, {
							switch(slot.efxindex,
								0x03, {			//porta to note
									if(slot.efxparam>0, {
										efxPortaSpeed[i]= slot.efxparam;
									});
									if(slot.period!=0, {
										if(slot.sample.notNil, {
											efxPortaPeriod[i]= slot.tunedPeriod;
										}, {
											efxPortaPeriod[i]= slot.period;
										});
									});
								},
								0x09, {			//sample offset
									trackLoopStart[i]= (slot.efxparam*256).min(trackLoopEnd[i]);
									trackSynth.set(\loopstart, trackLoopStart[i]);
								},
								0x0B, {			//jump to pattern
									index= slot.efxparam;
									row= -1;
								},
								0x0C, {			//set volume
									trackVolume[i]= slot.efxparam;
								},
								0x0D, {			//pattern break
									if(slot.efxparam>63, {
										row= -1;
									}, {
										row= slot.efxparam-1;
									});
									if(efxPatternBreakFlag.not, {
										index= index+1;
										efxPatternBreakFlag= true;
									});
								},
								0x0E, {
									switch(slot.efxparam.rightShift(4),
										0x0, {	//set filter
											filters[i].set(\mix, (slot.efxparam&0xF).min(1));
										},
										0x1, {	//fine porta up
											trackPeriod[i]= trackPeriod[i]-(slot.efxparam&0xF);
											trackSynth.set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
										},
										0x2, {	//fine porta down
											trackPeriod[i]= trackPeriod[i]+(slot.efxparam&0xF);
											trackSynth.set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
										},
										0x3, {	//glissando control
											efxGlissandoFlag= slot.efxparam&0xF>=1;
										},
										0x4, {	//vibrato waveform
											efxVibratoWave= slot.efxparam&0xF;
										},
										0x5, {	//set finetune
											if(slot.sample.notNil, {
												if(slot.efxparam&0xF>7, {slot.sample.tuning= slot.efxparam&0xF-16;
												}, {slot.sample.tuning= slot.efxparam&0xF;
												});
											});
										},
										0x6, {	//pattern loop
											if(slot.efxparam&0xF==0, {
												efxPatternLoopRow= row;
											}, {
												if(efxPatternLoopCnt==0, {efxPatternLoopCnt= slot.efxparam&0xF;
												}, {efxPatternLoopCnt= efxPatternLoopCnt-1;
												});
												if(efxPatternLoopCnt>0, {row= efxPatternLoopRow-1;
												});
											});
										},
										0x7, {	//tremolo waveform
											efxTremoloWave= slot.efxparam&0xF;
										},
										0x8, {	//16 pos panning
											trackSynth.set(\pan, slot.efxparam&0xF/7.5-1);
										},
										0xA, {	//fine volslide up
											trackVolume[i]= trackVolume[i]+(slot.efxparam&0xF);
										},
										0xB, {	//fine volslide up
											trackVolume[i]= trackVolume[i]-(slot.efxparam&0xF);
										},
										0xD, {	//delay note
											efxDelayNoteFlag[i]= true;
										},
										0xE, {	//delay pattern
											efxDelayPattern= slot.efxparam&0xF;
										}
										//0xF, {}	//invert loop -unused
									)
								},
								0x0F, {			//set speed
									if(slot.efxparam<0x20, {
										speed= slot.efxparam;
									}, {
										speed= 6;
										clk.tempo_(slot.efxparam/60);
									});
								}
							);
						});
						
						if(muteTracks.includes(i).not, {
							trackSynth.set(\vol, trackVolume[i].clip(0, 64)/64);
							
							if(slot.period>0, {	//trigger synth if note given
								if(slot.efxindex!=3 and:{slot.efxindex!=5}, {
									if(efxDelayNoteFlag[i].not, {
										if(efxDelayPattern==0, {
											trackSynth.set(\t_trig, 1);
										}, {
											efxDelayPattern= efxDelayPattern-1;
										});
									});
								});
							});
						});
					});							//end makeBundle
				};
				
				//--position updates
				if(row==63, {						//check if time to jump in order
					index= index+1;
				});
				row= (row+1)%64;					//jump to next row
				tick= 1;
			
			}, {									//update per tick (except tick 0)
				
				//--tick based effects
				tracks.do{|slot, i|				//for each track (usually 4)
					if(muteTracks.includes(i).not, {
						switch(slot.efxindex,
							0x00, {				//arpeggio
								if(slot.efxparam>0, {
									switch(tick-1%3,
										0, {
											synths[i].set(\rate, (RedMOD.period2note(trackPeriod[i])+(slot.efxparam&15)).midiratio);
										},
										1, {
											synths[i].set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
										},
										2, {
											synths[i].set(\rate, (RedMOD.period2note(trackPeriod[i])+slot.efxparam.rightShift(4)).midiratio);
										}
									);
								});
							},
							0x01, {				//porta up
								trackPeriod[i]= (trackPeriod[i]-slot.efxparam).clip(113, 907);
								synths[i].set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
							},
							0x02, {				//porta down
								trackPeriod[i]= (trackPeriod[i]+slot.efxparam).clip(113, 907);
								synths[i].set(\rate, RedMOD.period2note(trackPeriod[i]).midiratio);
							},
							0x03, {				//porta to note
								funcPorta.value(slot, i);
							},
							0x04, {				//vibrato
								if(tick==1, {
									if(slot.period>0 and:{efxVibratoWave<4}, {
										efxVibratoCount[i]= 0;
									});
									if(slot.efxparam!=0, {
										efxVibratoSpeed[i]= slot.efxparam.rightShift(4);
										efxVibratoDepth[i]= slot.efxparam&0x0F;
									});
								});
								funcVibrato.value(slot, i);
							},
							0x05, {				//porta + vol slide
								funcPorta.value(slot, i);
								funcVolumeSlide.value(slot, i);
							},
							0x06, {				//vibrato + vol slide
								if(tick==1, {
									if(slot.period>0 and:{efxVibratoWave<4}, {
										efxVibratoCount[i]= 0;
									});
								});
								funcVibrato.value(slot, i);
								funcVolumeSlide.value(slot, i);
							},
							0x07, {				//tremolo
								if(tick==1, {
									if(slot.period>0 and:{efxTremoloWave<4}, {
										efxTremoloCount[i]= 0;
									});
									if(slot.efxparam!=0, {
										efxTremoloSpeed[i]= slot.efxparam.rightShift(4);
										efxTremoloDepth[i]= slot.efxparam&15/4;
									});
								});
								switch(efxTremoloWave%4,
									0, {			//tremolo sine wave
										trackVolume[i]= (trackVolume[i]
											+(sin(efxTremoloCount[i]%64/64*pi)*efxTremoloDepth[i])
										).clip(0, 64);
										efxTremoloCount[i]= efxTremoloCount[i]+efxTremoloSpeed[i];
									},
									1, {			//tremolo ramp down
										trackVolume[i]= (trackVolume[i]
											+((efxTremoloCount[i]%64/64*2-1).neg*efxTremoloDepth[i])
										).clip(0, 64);
										efxTremoloCount[i]= efxTremoloCount[i]+efxTremoloSpeed[i];
									},
									2, {			//tremolo square wave
										trackVolume[i]= (trackVolume[i]
											+(efxTremoloCount[i].div(efxTremoloSpeed[i])%2*2-1*efxTremoloDepth[i])
										).clip(0, 64);
										efxTremoloCount[i]= efxTremoloCount[i]+1;
									},
									3, {
										trackVolume[i]= (trackVolume[i]+efxTremoloDepth[i].rand2).clip(0, 64);
									}
								);
								synths[i].set(\vol, trackVolume[i]/64);
							},
							0x08, {				//pan
								synths[i].set(\pan, slot.efxparam/0x80*2-1);
							},
							0x0A, {				//volume slide
								funcVolumeSlide.value(slot, i);
							},
							0x0E, {
								switch(slot.efxparam.rightShift(4),
									0x9, {		//retrig note
										if(tick-1%(slot.efxparam&0xF)==0, {
											synths[i].set(\t_trig, 1);
										});
									},
									0xC, {		//cut note
										if(tick-1==(slot.efxparam&0xF), {
											synths[i].set(\vol, 0);
										});
									},
									0xD, {		//delay note
										if(tick-1==(slot.efxparam&0xF), {
											synths[i].set(\t_trig, 1);
										});
									}
								)
							}
						);
					});
				};
				tick= tick+1%(customSpeed ? speed);
			});
			(1/4/6).wait;
		});
		1.wait;									//wait 1 beat before freeing synths
		synths.do{|x| x.free};
		filters.do{|x| x.free};
	}
}

//--a type of event class
RedMODSlot {
	var <>sample, <>period, <>efxindex, <>efxparam;
	*new {|sample, period, efxindex, efxparam|
		^super.newCopyArgs(sample, period, efxindex, efxparam)
	}
	note {										//amiga period as midi -12 to +23
		if(period>0, {^RedMOD.period2note(period)});
		^nil
	}
	tunedNote {^this.note+(sample.tuning/8)}			//midi note + sample finetuning
	tunedPeriod {^RedMOD.note2period(this.tunedNote)}	//amiga period + sample finetuning
}

//--a collection of 64 events
RedMODPattern {
	var <>slots;
	*new {|slots|
		^super.newCopyArgs(slots)
	}
	//ripPattern {}		//todo!!!
}

//--a type of buffer class
RedMODEmptySample {
	var <>name, <>tuning, <>volume, <>loopstart, <>looplen, <sampleRate,
		<>data, <buffer;
	*new {|name, tuning, volume, loopstart, looplen, sampleRate|
		^super.newCopyArgs(name, tuning, volume, loopstart, looplen, sampleRate)
	}
	empty {^true}
}
RedMODSample : RedMODEmptySample {
	empty {^false}
	load {|server, action|
		if(buffer.notNil, {this.free});
		buffer= Buffer.loadCollection(server, data, 1, action);
	}
	free {buffer.free}
	length {^buffer.numFrames ? 0}
	loop {^looplen>2}								//boolean
	loopend {
		if(this.loop, {^(looplen+loopstart).min(this.length)});
		^this.length
	}
	
	//--convert & load data into a float buffer
	ripSample {|server, action, interpolate= false|
		var copyBuf= data.copy;
		^this.prRip(server, action, interpolate, 1, copyBuf)
	}
	//--convert & load data into a float buffer - keeping looppoints, volume & tuning
	ripSample2 {|server, action, interpolate= false|
		var copyBuf= data.copy;
		var finetune= 2.pow(tuning/12/8);			//find finetuning rate
		if(looplen>2, {							//cut out looppoints if supposed to loop
			copyBuf= copyBuf.copyRange(loopstart, this.loopend);
		});
		if(volume!=64, {							//scale amp if volume given
			copyBuf= copyBuf*(volume/64);
		});
		^this.prRip(server, action, interpolate, finetune, copyBuf)
	}
	
	//--import soundfile into sample object via a buffer
	//  remember to manually update name, tuning, volume, loopstart, looplen
	importSoundFile {|path, server, action|
		server= server ? Server.default;
		if(server.serverRunning.not, {
			"RedMOD: boot server to import soundfile".error;
			^this
		});
		Routine.run{
			var halt= Condition.new;
			var file, buf;
			if(File.exists(path), {
				file= SoundFile.openRead(path);
				if(file.numChannels>1, {
					"RedMOD: multichannel soundfile - only first channel used".warn;
				});
				file.close;
				buf= Buffer.readChannel(server, path.standardizePath, channels:[0]);
				server.sync(halt);
				this.prImport(buf, action);
			}, {
				"RedMOD: file not found".error;
			});
		}
	}
	
	//--import buffer into sample object
	//  remember to manually update name, tuning, volume, loopstart, looplen
	importBuffer {|buffer, action|
		if(buffer.server.serverRunning.not, {
			"RedMOD: boot server to import buffer".error;
			^this
		});
		this.prImport(buffer, action);
	}
	
	//--private
	prRip {|server, action, interpolate, finetune, copyBuf|
		var bufsize;
		server= server ? Server.default;
		if(server.serverRunning.not, {
			"RedMOD: boot server to rip".error;
			^this
		});
		bufsize= (copyBuf.size*server.sampleRate*finetune/sampleRate).round;
		if(interpolate, {
			copyBuf= copyBuf.resamp1(bufsize)
		}, {
			copyBuf= copyBuf.resamp0(bufsize)
		});
		^Buffer.loadCollection(server, copyBuf, 1, action)
	}
	prImport {|buf, action|
		buf.loadToFloatArray(0, -1, {|arr|
			data= arr;
			sampleRate= buf.sampleRate;
			loopstart= 0;
			if(this.loop, {looplen= data.size});
			this.load(buf.server, action);
		});
	}
}
